/*
   $Id: CommandAlign.cpp 705 2013-01-25 10:42:19Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <string>
using namespace std;

#include "Bank.h"
#include "IndexTree.h"
#include "Comparison.h"
#include "utils.h"
#include "CommandAlign.h"

#ifdef _OPENMP
	#include <omp.h>
	#include "time.h"
#endif

#ifndef __LP64__
    #error "GLINT MAY BE COMPILED ONLY ON A 64 BITS PLATFORM"
#endif

extern RealLogger* logger;

/*////////////// Main ///////////////*/
int CommandAlignMap::Main(int argc,char* argv[])
{
	ComparisonBase* compare_ptr;

	try
	{
		Parameters& parameters = Parameters::Instance();
		ParametersInitializer& prm_init = _GetInitializer();
		prm_init.Init(argc, argv, *this);
		
		if( !parameters.IsSilent() ) 
		{
			PrintHeader();
		}
		
		#ifdef _OPENMP
		if (parameters.GetNbThreads() != 0)
		{
			omp_set_num_threads(parameters.GetNbThreads());
		}
		if ( !parameters.IsSilent() ) 
		{
			cerr << "using OPENMP parallelization directives - ";
			cerr << "at most " << omp_get_max_threads() << " threads will be running\n";
		}
		ThreadsTimeRecorder::Init();
		#endif

		Mask::Init();
		Bank::Init();
		
	    /* Initializing logger */
	    string log_filename;
	    string query_filename = parameters.GetQueryFilename().GetName();
	    if (query_filename == "-")
	    	log_filename="glint_pipe.log";
	    else
	    	log_filename=query_filename + ".log";
	    
	    auto_ptr<RealLogger> logger_auto = auto_ptr<RealLogger>(new RealLogger(log_filename,parameters.GetLogLevel(),LOGGER_OPT_AUTOFLUSH));
	    logger = logger_auto.get();
	    
	    logger->SetLogFormat(no_stamp);

	    Bank* bank_ptr = Bank::InstancePtr();
	    
	    // The bank is not indexed
	    if ( parameters.IsIndexMode() )
	    {
			MemoryPool::Init();
			IndexTree::Init();
			
			// We want to save the index
			if ( parameters.IsSaveIndexMode() )
			{
				if ( !parameters.IsSilent() )
				{
					cerr << "Indexing the bank and saving the index \n";
				}
			    bank_ptr -> Index(false);
				prm_init.Init(argc, argv, *this);
				
			    //bank_ptr -> Remove();
			    Bank::ReInit();
			    bank_ptr = Bank::InstancePtr();
			}
			
			// No use saving the index (default): just keep them in memory
			else
			{
				if ( !parameters.IsSilent() )
				{
					cerr << "Indexing the bank and keeping the index in memory \n";
				}
				bank_ptr -> Index(true);
			}
			
			// removing data structures no more useful
			Singleton<IndexTree>::Remove();
			Singleton<MemoryPool>::Remove();
		}
		
		// TODO Remove : parameters.GetDeepestLevel() dprecated
		// check some parameters relative to the mask 
		/*Mask& mask = Mask::Instance();
		if ( mask.GetLeafLevel() < parameters.GetDeepestLevel() )
		{
			ostringstream err;
			err << "ERROR - The switch -d (" << (int) parameters.GetDeepestLevel() << ") should NOT be > mask leaf level (" << (int) mask.GetLeafLevel() << ") - ";
			err << "here is the mask - " << mask.GetString();
			throw (runtime_error(err.str()) );
		}*/
		
		CompactTree::Init();
		AlignConstants::Init();
		BitCount::Init();
		ScoreMatrix::Init();
		MatchConstants::Init();
		ReverseComplement::Init();

        if (parameters.IsSelf() && !parameters.IsSilent())
            cerr << "Self comparison has been asked !\n";

#ifdef DEBUGLOG
        cerr << "Bank     : " << parameters.GetBankFilename()  << endl;
        cerr << "Request  : " << parameters.GetQueryFilename() << endl;
        cerr << "Output   : " << parameters.GetOutputFilename()<< endl;
#endif

		if ( parameters.IsSelf() )
		{
			compare_ptr = new ComparisonSelf();
		}
		else if ( parameters.IsGenome() )
		{
			compare_ptr = new ComparisonGenome();
		}
		else
		{
			compare_ptr = new Comparison();
		}	
		
		compare_ptr -> Run();
		parameters.Warn();
		
		delete compare_ptr;

#ifdef _OPENMP
		if ( !parameters.IsSilent() ) 
		{
			//ThreadsTimeRecorder& ttr = ThreadsTimeRecorder::Instance();
			//cerr << ttr << '\n';
		}
#endif
			
	}
   	catch ( exception& e)
	{	
		cout << e.what() << endl;
		return 1;
	};
	
	return 0;
}


/*Utilisation du programme */
void CommandAlign::Usage() const
{
	//PrintHeader();
	cerr << "\n";
	cerr << "Usage: (1) glint align [options] <bank> <query> \n";
	cerr << "       (2) glint align [options] <bank>  - \n\n";		
	
	cerr << "\t(1-2) bank  multifasta\n";
	cerr << "\t\tsequence should have been indexed first by the index command (however, if index files not found, the index is computed and kept in memory)\n";
	cerr << "\t(1) query multifasta (see below for other formats) \n";
	cerr << "\t(2) the query is read from stdin, the format must be fasta or multifasta, uncompressed\n";
	cerr << "\n";
	cerr << "\tfiles may be compressed using Z,GZ,BZ2,LZMA,XZ algorithms\n";	
	
	cerr << "Options:   \n";
	cerr << "\t-G   query.fasta must have been previously indexed by glint index - Faster for complete genome comparisons\n";
	cerr << "\t-f   fast output without sorting hsps\n";
	cerr << "\t--save-index  if the index files are not found and --save-index is specified, the index is saved to disk (equivalent to performing the index command first, see index options)\n";
	cerr << "\n";	
	cerr << ComplexityManagementUsage();
	cerr << ControllingAlignmentUsage();
	cerr << InputFormatUsage();	
	cerr << OutputFormatUsage();
	cerr << OutputCompressionUsage();
	cerr << GeneralUsage();
	
	cerr << "Usage(2): glint align [options] --self bank \n\n";
	cerr << "\t -S, --self  self comparison, the bank is compared against itself, diagonal avoided and only upper triangle of the dotplot is computed (query pos < subject pos) \n";  
	cerr << "Options:   \n";
    cerr << "\t\t --tandem INT  maximum size allowed for tandem alignment\n";
    
}

/*Utilisation du programme */
void CommandSelf::Usage() const
{
	//PrintHeader();	
	cerr << "\nUsage: glint self [options] <sequence> \n\n";
			
	cerr << "\tsequence multifasta\n";
	cerr << "\t\tsequence should have been indexed first by the index command (however, if index files not found, the index is computed and kept in memory)\n";
	cerr << "\t\tfile may be compressed using Z,GZ,BZ2,LZMA,XZ algorithms\n";	
	
	cerr << "Options:   \n";
	cerr << "\t --tandem INT  maximum size allowed for tandem alignment\n";
	cerr << "\t-f   fast output without sorting hsps\n";
	cerr << "\t--save-index  if the index files are not found and --save-index is specified, the index is saved to disk (equivalent to performing the index command first, see index options)\n";
	cerr << "\n";	
	// retire par MANU parce que le switch --upper n'a pas de sens ici
	//	cerr << ComplexityManagementUsage();
	cerr << "\tcomplexity managment\n";
	cerr << "\t\t --no-lc-filtering disable low complexity filtering (ignored if data already indexed)\n";
    cerr << "\t\t --strict  lower case letter in the query sequence will be considered as non reliable nucleotides (=> they always lead to a mismatch)\n";
	

	cerr << ControllingAlignmentUsage();
	cerr << InputFormatUsage();	
	cerr << OutputFormatUsage();
	cerr << OutputCompressionUsage();
	cerr << GeneralUsage();
 
}

/*Utilisation du programme */
void CommandMap::Usage() const
{
	//PrintHeader();
	cerr << "\nUsage: glint map [options] <bank> <query> \n\n";

	cerr << "\tbank  multifasta\n";
	cerr << "\t\tsequence should have been indexed first by the index command (however, if index files not found, the index is computed and kept in memory)\n";
	cerr << "\tquery multifasta, fastq, sam, bam  \n";
	cerr << "\n";
	cerr << "\tfiles may be compressed using Z,GZ,BZ2,LZMA,XZ algorithms\n";	
	cerr << "Options:   \n";
	cerr << "\t--save-index  if the index files are not found and --save-index is specified, the index is saved to disk (equivalent to performing the index command first, see index options)\n";
	cerr << "\n";	
	cerr << MultiprocessorUsage();
	cerr << ComplexityManagementUsage();
	cerr << ControllingAlignmentUsage();
	cerr << InputFormatUsage();	
	cerr << OutputFormatUsage();
	cerr << OutputCompressionUsage();	
	cerr << GeneralUsage();
	
}

/*Utilisation du programme */
void CommandMappe::Usage() const
{
	//PrintHeader();
	cerr << "\n";
	cerr << "Usage: (1) glint mappe [options] <bank> <query> \n";
	cerr << "       (2) glint mappe [options] <bank> <query_fwd>,<query_bwd> \n\n";
	
	cerr << "\t(1-2) bank  multifasta\n";
	cerr << "\t\tbank should have been indexed by the index command (if index files not found, the index is computed and kept in memory)\n";
	cerr << "\t(1) query sam, bam  \n";
	cerr << "\t\tusefull when the reads are available only through a sam(bam) file\n";
	cerr << "\t\tmate-pair sequences in the sam(bam) must share the same name\n";
	cerr << "\t(2) query_fwd and query_bwd  multifasta or fastq\n";
	cerr << "\t\tthe query_fwd and the query_bwd should be parallel files: same number of sequences, same Id for the sequences and in the same order\n";
	cerr << "\n";
	cerr << "\tfiles may be compressed using Z,GZ,BZ2,LZMA,XZ algorithms\n";	
	cerr << "Options:   \n";
	cerr << " \t\t--epcr mapping primers to the bank. Just as --pair-end but with additional constraints (only FR orientation is allowed: L-> <-R)\n";
	cerr << " \t\t--mate-maxdist INT maximum allowed distance between mate-pairs ["<<DEFAULT_MAX_DISTMATE<< "]\n";
	cerr << MultiprocessorUsage();
	cerr << ComplexityManagementUsage();
	cerr << ControllingAlignmentUsage();
	cerr << InputFormatUsage();	
	cerr << OutputFormatUsage();
	cerr << OutputCompressionUsage();	
	cerr << GeneralUsage();
	
}

string CommandAlignMap::ControllingAlignmentUsage() const
{	
	ostringstream out;
	
	out << "\tcontrolling the alignment\n";
	out << "\t\t -s INT minimum score ro report an alignment [" << DEFAULT_CUTOFF_HSP <<  std::string("]\n");
	out << "\t\t -c INT minimum score to trigger an ungapped alignment [" << DEFAULT_CUTOFF_SLIDE << "] \n";
	out << "\t\t -X INT minimum score to trigger a gapped alignment (dynamic programming) [" << DEFAULT_CUTOFF_DYN << "]\n";
	out << "\t\t\ts and x should be positive integers with s > X\n";
	out << "\t\t -w INT: where w is the starting point at which the seeds are reported (similar to -W in blast) [" << DEFAULT_LOWER_LEVEL << "]\n";
	out << "\t\t\tw should be a positive integer in the range 1-8, with w <= mask leaf level\n";
	out << "\t\t --step INT stepping value for the window on the query sequence in the process of identifying seeds [" << DEFAULT_STEP << "]\n";
	out << "\t\t --no-gap  do not trigger gapped alignment \n";	
	out << "\tscoring sheme\n";
	out << "\t\t --match=INT     reward for a match ["<<ALIGN_CONSTANTS_MATCH<<"]\n";
    out << "\t\t --mismatch=INT  reward (penality) for a mismatch ["<<ALIGN_CONSTANTS_MISMATCH<<"]\n";
	out << "\t\t --gapop=INT     reward (penality) for a gap opening ["<<ALIGN_CONSTANTS_GAPOPEN<<"]\n";
    out << "\t\t --gapext=INT    reward( penality) for a gap extension ["<<ALIGN_CONSTANTS_GAPEXTEN<<"]\n";	
	
	out << "\tfiltering the output\n";
	out << "\t\t --best-score for each hit, keep only the hsp with best score\n";
	out << "\t\t --mmis  INT maximum number of mismatches allowed [none]\n";
    out << "\t\t --mgap  INT maximum number of gaps allowed [none]\n";
    out << "\t\t --imin  INT minimum identity of the hsp [none]\n"; 
    out << "\t\t --lmin  INT minimum length of the hsp [none]\n"; 
    out << "\t\t --srmin INT  minimum score, expressed in terms of percentage of the maximum possible score given the query [none]\n";
    out << "\t\t --lrmin INT  minimum length, expressed in terms of percentage of the maximum possible length given the query [none]\n";
		
	return out.str();
}

string CommandAlignMap::InputFormatUsage() const
{	
	ostringstream out;
	
	out << "\tinput format\n";
	out << "\t\t --input-format fa,fastq,sam,bam  input format [" << DEFAULT_GLINT_INPUT_FORMAT_DSP << "]\n";
	out << "\t\t\tfa   : the fasta format\n";
	out << "\t\t\tfastq: the fastq format\n";
	out << "\t\t\tsam  : the sam format\n";
	out << "\t\t\tbam  : the bam format [NOTE - Compression is handled by the bam format]\n";	
		
	return out.str();	
}

string CommandAlignMap::OutputFormatUsage() const
{	
	ostringstream out;
	
	out << "\toutput format\n";
	out << "\t\t -m 1-3,8-9 : output format [" << DEFAULT_GLINT_OUTPUT_FORMAT_DSP << "]\n";
    out << "\t\t --output-format 1-3,8-9 : output format (same as -m)\n";
	out << "\t\t\t1    complete alignement\n";
	out << "\t\t\t2    a tabular output \n";
	out << "\t\t\t3    shortest edit script \n";
	out << "\t\t\t8    the blastall -m 8 ouput format \n";
	out << "\t\t\t9    the blastall -m 9 ouput format \n";			
	
	return out.str();	
}

string CommandAlignMap::OutputCompressionUsage() const
{	
	ostringstream out;
	
	out << "\tredirecting and compressing the output\n";
	out << "\t\t -o filename, --output-file filename : output file (default stdout)\n";
	out << "\t\t\t when using the -o switch, the default is to use THE SAME compression algorithm as the input file\n";
	out << "\t\t\t this can be changed with the --compress switch:\n";
	out << "\t\t --compress Z|GZ|BZ2|LZMA|XZ Force one of those compression algorithms\n";
	out << "\t\t --compress any              Use the most efficient compression algorithm (depends on the programs installed on the host)\n";	
	
	return out.str();		
	
}

string CommandMap::OutputFormatUsage() const
{	
	ostringstream out;
	
	out << "\toutput format\n";
	out << "\t\t -m 1-3,8-9 : output format [" << DEFAULT_GLINT_OUTPUT_FORMAT_DSP << "]\n";
    out << "\t\t --output-format 1-3,8-9 : output format (same as -m)\n";
	out << "\t\t\t1    complete alignement\n";
	out << "\t\t\t2    a tabular output \n";
	out << "\t\t\t3    shortest edit script \n";
	out << "\t\t\t8    the blastall -m 8 ouput format \n";
	out << "\t\t\t9    the blastall -m 9 ouput format \n";			
	out << "\t\t\tsam  : the sam format [NOTE --compress is IGNORED with the sam/bam formats] \n";
	out << "\t\t\tbam  : the bam format\n";
	out << "\t\t\tsamext: the sam format with the MD optional field (storing the whole alignment)\n";
	out << "\t\t\tbamext: the bam format with the MD optional field (storing the whole alignment)\n";
	out << "\t\t\tbed  : the bed format\n";
	
	return out.str();	
}

string CommandMap::MultiprocessorUsage() const
{	
	ostringstream out;
	
	out << "\tmultiprocessor\n";
	#ifdef _OPENMP
	out << "\t\t -a auto : set the number of threads appropriate to your hardware\n";
	out << "\t\t -a n    : specify the number of threads \n";
	#else
	out << "\t\t Please recompile glint to get multiprocessor capabilities\n";
	#endif
	
	return out.str();	
}


string CommandMappe::ControllingAlignmentUsage() const
{	
	ostringstream out;
		
	out << "\tcontrolling the alignment\n";
	out << "\n\t   in mappe mode, except --best-score which only keeps the aligned pairs with the best score,\n";
	out << "\t   all the other options below affect the alignement for each read of a pair independently\n\n";
	out << "\t\t -s INT minimum score ro report an alignment [" << DEFAULT_CUTOFF_HSP <<  std::string("]\n");
	out << "\t\t -c INT minimum score to trigger an ungapped alignment [" << DEFAULT_CUTOFF_SLIDE << "] \n";
	out << "\t\t -X INT minimum score to trigger a gapped alignment (dynamic programming) [" << DEFAULT_CUTOFF_DYN << "]\n";
	out << "\t\t\ts and x should be positive integers with s > X\n";
	out << "\t\t -w INT: where w is the starting point at which the seeds are reported (similar to -W in blast) [" << DEFAULT_LOWER_LEVEL << "]\n";
	out << "\t\t\tw should be a positive integer in the range 1-8, with w <= mask leaf level\n";
	out << "\t\t --step INT stepping value for the window on the query sequence in the process of identifying seeds [" << DEFAULT_STEP << "]\n";
	out << "\t\t --no-gap  do not trigger gapped alignment \n";	
	out << "\tscoring sheme\n";
	out << "\t\t --match=INT     reward for a match ["<<ALIGN_CONSTANTS_MATCH<<"]\n";
    out << "\t\t --mismatch=INT  reward (penality) for a mismatch ["<<ALIGN_CONSTANTS_MISMATCH<<"]\n";
	out << "\t\t --gapop=INT     reward (penality) for a gap opening ["<<ALIGN_CONSTANTS_GAPOPEN<<"]\n";
    out << "\t\t --gapext=INT    reward( penality) for a gap extension ["<<ALIGN_CONSTANTS_GAPEXTEN<<"]\n";	
	
	out << "\tfiltering the output\n";
	out << "\t\t --best-score for each hit, keep only the hsp with best score\n";
	out << "\t\t --mmis  INT maximum number of mismatches allowed [none]\n";
    out << "\t\t --mgap  INT maximum number of gaps allowed [none]\n";
    out << "\t\t --imin  INT minimum identity of the hsp [none]\n"; 
    out << "\t\t --lmin  INT minimum length of the hsp [none]\n"; 
    out << "\t\t --srmin FLOAT  minimum score, expressed in terms of percentage of the maximum possible score given the query [none]\n";
    out << "\t\t --lrmin FLOAT  minimum length, expressed in terms of percentage of the maximum possible length given the query [none]\n";
		
	return out.str();
}


string CommandMappe::OutputFormatUsage() const
{	
	ostringstream out;
	
	out << "\toutput format\n";
	out << "\t\t -m sam,bam,samext,bamext,bed : output format [" << DEFAULT_GLINT_OUTPUT_FORMAT_DSP << "]\n";
    out << "\t\t --output-format sam,bam,samext,bamext,bed : output format (same as -m)\n";
	out << "\t\t\tsam  : the sam format [NOTE --compress is IGNORED with the sam/bam formats] \n";
	out << "\t\t\tbam  : the bam format\n";
	out << "\t\t\tsamext: the sam format with the MD optional field (storing the whole alignment)\n";
	out << "\t\t\tbamext: the bam format with the MD optional field (storing the whole alignment)\n";
	out << "\t\t\tbed  : the bed format, bedpe here in pair-end mode\n";
	
	return out.str();	
}



