/*
   $Id: CommandConvert.cpp 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <string>
#include <memory>
using namespace std;

#include "ptypes.h"
#include "QueryBuffer.h"
#include "QueryBufferDouble.h"

#include "Display.h"
#include "DisplayM1.h"
#include "DisplaySam.h"

#include "ParametersConvert.h"
#include "CommandConvert.h"

#ifndef __LP64__
    #error "GLINT MAY BE COMPILED ONLY ON A 64 BITS PLATFORM"
#endif

void Usage();
extern RealLogger* logger;

auto_ptr<QueryBuffer> InitQueryBuffer (const string& seq_file_name, const string& mate_file_name, CompressionAlgorithm compression, int input_format, int output_format, bool pair_end)
{
	//bool quality = (output_format == OUTPUT_FORMAT_FASTQ);
	bool quality = true;
	auto_ptr<QueryBuffer> q(NULL);
	if ( pair_end==true )
	{
		switch (input_format)
		{
			case INPUT_FORMAT_FA   : q.reset ( new QueryBufferFastaDouble(seq_file_name, mate_file_name, compression ) ); break;
			case INPUT_FORMAT_FASTQ: q.reset ( new QueryBufferFastqDouble(seq_file_name, mate_file_name, compression ) ); break;
			case INPUT_FORMAT_SAM: throw(runtime_error("ERROR - INPUT FORMAT SAM NOT SUPPORTED BY GLINTCONVERT IN PAIR END - TRY BAM"));
			case INPUT_FORMAT_BAM: q.reset ( new QueryBufferSamDouble(seq_file_name, quality, true ) ); break;
			default: throw(runtime_error("ERROR - INPUT FORMAT NOT SUPPORTED BY GLINTCONVERT"));
		}
	}
	else
	{
		switch (input_format)
		{
			case INPUT_FORMAT_FA   : q.reset ( new QueryBufferFasta(seq_file_name, compression ) ); break;
			case INPUT_FORMAT_FASTQ: q.reset ( new QueryBufferFasta(seq_file_name, compression ) ); break;
			case INPUT_FORMAT_SAM  : q.reset ( new QueryBufferSam(seq_file_name, false, quality, true ) ); break;
			case INPUT_FORMAT_BAM  : q.reset ( new QueryBufferSam(seq_file_name, true,  quality, true ) ); break;
			default: throw(runtime_error("ERROR - INPUT FORMAT NOT SUPPORTED BY GLINTCONVERT"));
		}
	}
	return q;
}

/**
 * \brief Initialize the display for the seq OR for the mate
 * 
 * \param output_format The output format
 * \param query The QueryBuffer to connect to the display
 * \param If true, we are creating the display for the mate
 */

auto_ptr<Display> InitDisplay (int output_format, const QueryBuffer* query, bool mate=false)
{
	auto_ptr<Display> d(NULL);
	switch (output_format)
	{
		case OUTPUT_FORMAT_M0:    d.reset( new DisplayM0() ); break;
		case OUTPUT_FORMAT_M1:    d.reset( new DisplayM1() ); break;
		case OUTPUT_FORMAT_M2:    d.reset( new DisplayM2(mate) ); break;
		case OUTPUT_FORMAT_M3:    d.reset( new DisplayM3(mate) ); break;
		case OUTPUT_FORMAT_M4:    d.reset( new DisplayM4(mate) ); break;
		case OUTPUT_FORMAT_M8:    d.reset( new DisplayM8(mate) ); break;
		case OUTPUT_FORMAT_M9:    d.reset( new DisplayM9(mate) ); break;
		case OUTPUT_FORMAT_FA:    d.reset( new DisplayFasta(mate) ); break;
		case OUTPUT_FORMAT_FASTQ: d.reset( new DisplayFastq(mate) ); break;
		case OUTPUT_FORMAT_BED  : d.reset( new DisplayBed() ); break;
		default: throw(runtime_error("ERROR - OUTPUT FORMAT NOT SUPPORTED BY GLINTCONVERT\n"));
	}
	d -> SetQuery(query);
    d -> DisplayTopHeader();
	return d;
}

Display* InitDisplay(Parameters& parameters, const QueryBuffer* query, auto_ptr<Display>& display_seq_aptr, auto_ptr<Display>& display_mate_aptr, size_t chunk_nb=0)
{
	if ( chunk_nb != 0 )
	{
		parameters.SetChunkNb( chunk_nb );
	}
	//display_seq_aptr = InitDisplay(parameters.GetOutputFormat(),query);	
	Display* display_mate_ptr = NULL;
	if ( parameters.IsPairEnd() == true)
	{		
		QueryBufferSamDouble* querydouble = (QueryBufferSamDouble*) query;
		display_seq_aptr = InitDisplay(parameters.GetOutputFormat(),&(querydouble->GetFwdQueryBuffer()));
		if (parameters.GetOutputBwdFilename())
		{
			display_mate_aptr= InitDisplay(parameters.GetOutputFormat(),&(querydouble->GetBwdQueryBuffer()),true);
			display_mate_ptr = display_mate_aptr.get();
		}
		///else
		//{
		//	display_mate_ptr = display_seq_aptr.get();
		//}
	}
	else
	{
		display_seq_aptr = InitDisplay(parameters.GetOutputFormat(),query);	
	}
		
	return display_mate_ptr;
}
	
/*////////////// Main ///////////////*/
int CommandConvert::Main(int argc,char* argv[])
{
	try
	{
		Parameters& parameters = Parameters::Instance();
		ParametersInitializer& prm_init = _GetInitializer();
		prm_init.Init(argc, argv, *this);
		ReverseComplement::Init();
		
		if ( !parameters.IsSilent() ) 
			PrintHeader();
		
	    string log_filename;
	    string query_filename            = parameters.GetQueryFilename().GetName();
	    string bwd_filename              = parameters.GetBwdFilename().GetName();
	    CompressionAlgorithm compression = parameters.GetQueryFilename().GetCompression();
	    
	    /* Initializing logger */
	    if (query_filename == "-")
	    	log_filename="glint_pipe.log";
	    else
	    	log_filename=query_filename + ".log";
	    
	    auto_ptr<RealLogger> logger_auto = auto_ptr<RealLogger>(new RealLogger(log_filename,parameters.GetLogLevel(),LOGGER_OPT_AUTOFLUSH));
	    logger = logger_auto.get();
	    
	    logger->SetLogFormat(no_stamp);

		bool pair_end = parameters.IsPairEnd();
		auto_ptr<QueryBuffer> query_aptr   = InitQueryBuffer(query_filename,bwd_filename,compression,parameters.GetInputFormat(),parameters.GetOutputFormat(),pair_end);
		
		//auto_ptr<Display>     display_aptr = InitDisplay(parameters.GetOutputFormat(),query_aptr.get());
		
		// Init One of Two display objects (if there are a seq and its mate in two different output files)
		auto_ptr<Display> display_seq_aptr(NULL);
		auto_ptr<Display> display_mate_aptr(NULL);
		Display* display_mate_ptr(NULL);
		uint64_t chunk_size = parameters.GetOutputChunkSize();
		uint64_t chunk_nb = (chunk_size!=0) ? 1 : 0;
		display_mate_ptr = InitDisplay(parameters,query_aptr.get(),display_seq_aptr,display_mate_aptr,chunk_nb);

		string last_query_name;
		Hsp h0;   // A fake hsp, may be useful to display something if GetHsp returns NULL
		uint64_t seq_nb   = 0;
		if ( pair_end == true )
		{
			QueryBufferDouble* query_ptr = (QueryBufferDouble*) query_aptr.get();
			while (query_aptr -> ReadNextSeq())
			{
				const Hsp* hs = query_ptr -> GetHsp();
				if ( hs==NULL)
				{
					hs = &h0;
				}
				const Hsp* hm = query_ptr -> GetHspMate();
				if ( hm==NULL)
				{
					hm = &h0;
				}

				if ( hs -> GetQueryName() != last_query_name )
				{
				    display_seq_aptr -> DisplayQueryInfo(hs -> GetQueryName());
				    display_seq_aptr -> DisplaySubjectInfo(hs -> GetSubjectName(), hs -> GetScore());
				    if ( display_seq_aptr.get() != display_mate_ptr )
				    {
					    if (display_mate_aptr.get() != NULL)
					    {
						    display_mate_aptr -> DisplayQueryInfo(hs -> GetQueryName());
						    display_mate_aptr -> DisplaySubjectInfo(hs -> GetSubjectName(), hs -> GetScore());
						}
					}
				    last_query_name = hs -> GetQueryName();
				}
				display_seq_aptr -> DisplayHsp(*hs);
				if (display_mate_aptr.get() != NULL)
				{
					display_mate_ptr -> DisplayHsp(*hm);
				}
			
				if ( chunk_size != 0 && ++seq_nb == chunk_size )
				{
					seq_nb = 0;
					chunk_nb++;
					display_mate_ptr = InitDisplay(parameters,query_aptr.get(),display_seq_aptr,display_mate_aptr,chunk_nb);
				}
			}
		}
		else
		{
			while (query_aptr -> ReadNextSeq()==true)
			{
				const Hsp* h = query_aptr -> GetHsp();
				if ( h==NULL)
				{
					h = &h0;
				}
				if ( h -> GetQueryName() != last_query_name )
				{
				    display_seq_aptr -> DisplayQueryInfo(h -> GetQueryName());
				    display_seq_aptr -> DisplaySubjectInfo(h -> GetSubjectName(), h -> GetScore());
				    last_query_name = h -> GetQueryName();
				}
				display_seq_aptr -> DisplayHsp(*h);
				if ( chunk_size != 0 && ++seq_nb == chunk_size )
				{
					seq_nb = 0;
					chunk_nb++;
					display_mate_ptr = InitDisplay(parameters,query_aptr.get(),display_seq_aptr,display_mate_aptr,chunk_nb);
				}
			}
		}

		parameters.Warn();

    }
   	catch ( exception& e)
	{	
		cout << e.what() << endl;
		return 1;
	};

	return 0;
}


/*Utilisation du programme */
void CommandConvert::Usage() const
{
	PrintHeader();
	cerr << "glint convert: Converting alignment or sequence file formats.\n\n";
	cerr << "Usage:  (1) glint convert --input-format [fasta|fastq|sam|bam] --output-format [fastq|fasta|bed|2|3|4|8|9] input [-o output]\n";
	cerr << "        (2) glint convert [options] '-' \n";
	cerr << "\tAutomatic detection of input file format is NOT specified, you have to specify the input format\n";
	cerr << "\tYou may convert fasta => fastq, fastq => fasta, sam/bam => any specified format\n";
	cerr << "\n\t\t========== SPECIFYING THE INPUT FORMAT  ============\n";
    cerr << "\t\t --input-format fastq,sam,bam : input format (default " << DEFAULT_GLINTCONVERT_INPUT_FORMAT_DSP << ")\n";
	cerr << "\t\t\tfasta: the fasta format\n";
	cerr << "\t\t\tfastq: the fastq format\n";
	cerr << "\t\t\tsam  : the sam/samext format\n";
	cerr << "\t\t\tbam  : the bam/bamext format [NOTE - Compression is handled by the bam format]\n";
	cerr << "\t\t --pair-end : Input and output data are paired. See under for the details\n";
	cerr << "\n\t\t========== FORMATTING THE OUTPUT        ============\n";
    cerr << "\t\t -m 1-3,8-9,fasta,fastq : output format (default m=" << DEFAULT_GLINTCONVERT_OUTPUT_FORMAT_DSP << ")\n";
    cerr << "\t\t --output-format 1-3,8-9,fa,fastq : output format (same as -m)\n";
	cerr << "\t\t\t1    : complete alignement\n";
	cerr << "\t\t\t2    : tabular output \n";
	cerr << "\t\t\t3    : shortest edit script \n";
	cerr << "\t\t\t8    :  the blastall -m 8 ouput format \n";
	cerr << "\t\t\t9    : the blastall -m 9 ouput format \n";
	cerr << "\t\t\tfasta: the fasta format\n";
	cerr << "\t\t\tfastq: the fastq format\n";
	cerr << "\t\t --chunk-size 1000: When there are more than 1000 sequences to write, a new output file is opened\n";
	cerr << "\t\t WARNING - The output file names should have the substring \"%chunk\" somewhere in their names, it will be replaced with a number !\n";
	cerr << "\n\t\t========== REDIRECTING AND COMPRESSING THE OUTPUT ============\n";
	cerr << "\t\t -o filename, --output-file filename : output file (default stdout)\n";
	cerr << "\t\t -o seq_filename,mate_filename : specify TWO files, one for the sequence, one for the mate. Works ONLY if --pair-end is specified\n";
	cerr << "\t\t When using the -o switch, the default is to use THE SAME compression algorithm as the input file\n";
	cerr << "\t\t This can be changed with the --compress switch:\n";
	cerr << "\t\t --compress Z|GZ|BZ2|LZMA|XZ Force one of those compression algorithms\n";
	cerr << "\t\t --compress any              Use the most efficient compression algorithm (depends on the programs installed on the host)\n";
}
