/*
   $Id: CommandIndex.cpp 666 2011-11-15 10:16:53Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

using namespace std;
#include <iostream>
#include <fstream>
#include <stdexcept>

#include "utils.h"
#include "Bank.h"
#include "memory.h"
#include "IndexTree.h"

#include "logger.h" 

#include "ParametersIndex.h"
#include "CommandIndex.h"

#ifndef __LP64__
    #error "GLINT MAY BE COMPILED ONLY ON A 64 BITS PLATFORM"
#endif

extern RealLogger* logger;

/*////////////// Main ///////////////*/
int CommandIndex::Main(int argc,char* argv[])
{
	try
	{
		Parameters& parameters = Parameters::Instance();
		ParametersInitializer& prm_init = _GetInitializer();
		prm_init.Init(argc, argv, *this);

		Mask::Init();
		Bank::Init();
		MemoryPool::Init();
		IndexTree::Init();
		
		Bank& bank = Bank::Instance();
		
	    string filename = bank.GetFileName().GetName();
		if( !parameters.IsSilent() ) 
		{
			PrintHeader();
			cerr << "Reading the sequence file " << filename << '\n';
		    cerr << "Mask = " <<  parameters.GetMskStr() << '\n';
		}
	    
	    // Initializing logger: we use an auto_ptr to be sure that the destructor is called at end of program
	    auto_ptr<RealLogger> logger_ptr = auto_ptr<RealLogger> (new RealLogger(filename + ".log",parameters.GetLogLevel(),LOGGER_OPT_AUTOFLUSH));
	    logger = logger_ptr.get();
	
		// Starting the indexation process
		bank.Index();
		
		return 0;
	}
	catch ( exception& e)
	{	
		cout << e.what() << endl;
	};
	return 0;
}

/*Utilisation du programme */
void CommandIndex::Usage() const
{
	//PrintHeader();
	//cerr << "glint index: Build the index of a nucleotide sequence to be used by glint compare, glint map, etc.\n\n";
	cerr << "\nUsage: glint index [options] <bank>\n\n";
	cerr << "\tbank nucleotide sequence to index (a chromosome for example) in multi-fasta format.\n";
	cerr << "\t          may be compressed using one of the following algorithms: Z|GZ|BZ2|LZMA|XZ\n";
	cerr << "\t          resulting files will be compressed using the SAME algorithm\n";
	cerr << "Options:\n";
	cerr << "\tindexing options\n";	
	cerr << "\t\t -N STRING   name of the bank (defaults to file name)\n";
	cerr << "\t\t -M STRING   mask structure (default " <<  DEFAULT_MASK << ", max 32 characters), may contain 1,..," << MAX_LEAF_LEVEL << " characters.\n";
	cerr << "\t\t --mask      same as -M\n";
	cerr << "\t\t -p INT      number of passes to create the index (default 1, the higher the less memory used, but the longest in time)\n";
    cerr << "\t\t             not really useful, the number of passes is automatically computed. The --verbose switch prints the memory consumption\n";
	cerr << "\t\t -n INT      maximum number of allowed occurences for a word otherwise deleted (default " << DEFAULT_MAX_COUNT<< ", max 256).\n";
	cerr << "\t\t --no-lc-filtering  disable low complexity filtering\n";
	cerr << GeneralUsage();

    #ifdef DEBUGLOG
	cerr << "\t\t -D T/TV Log some information about IndexTree (cf. IndexTree.cpp)\n";
    #endif
}
