/*
   $Id: Parameters.h 680 2012-01-30 14:42:53Z tfaraut $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef PARAMETERS_H__
#define PARAMETERS_H__

#include "ptypes.h"
#include <map>
#include <vector>
#include "utils.h"
#include "SimpleOpt.h"

//enum { GLINTINDEX, GLINT, GLINTSTAT, GLINTCHAIN, GLINTCONVERT, GLINTFREQ };

/**
 * \brief This class is used to support Data compression for the input or output files
 *        It is a singleton
 *        The methods getExtension, GetProgram return the correct file extension or program, given the compression algorithm
 *        GetInfo returns a CompressionInfo::CmpInf structure, given a file extension
 *        SearchBestCompression looks for the programs installed on the host, and returns the most performant CompressionAlgorithm
 */

string ToUpper(const string& s);
string ToLower(const string& s);
class CompressionInfo: public Singleton<CompressionInfo> 
{
	public:
		friend class Singleton<CompressionInfo>;
		
		struct CmpInf {
			CompressionAlgorithm algorithm;
			string extension;
			string program;
		};
		
		class IsExtension {
			public:
				IsExtension(const string& e):__e(e){};
				bool operator() (const CmpInf& c) { return (ToLower(c.extension) == ToLower(__e)); };
			private:
				string __e;
		};
		
			
		CompressionInfo();
		CompressionAlgorithm SearchBestCompression() const;
		const string& GetExtension (CompressionAlgorithm a) const { return __cmp_infos.at(a).extension ; };
		const string& GetProgram   (CompressionAlgorithm a) const { return __cmp_infos.at(a).program ; };
		const CmpInf& GetInfo      (const string& extension) const
		{
			IsExtension is_ext(extension);
			vector<CmpInf>::const_iterator found = find_if(__cmp_infos.begin(),__cmp_infos.end(),is_ext);
			if ( found == __cmp_infos.end())
			{
				return *__cmp_infos.begin();
			}
			else
			{
				return *found;
			}
		};
		
	private:
		vector<CmpInf> __cmp_infos;
};

/**
 * \brief This class is used to store a file name, together with its compression algorithm and extension
 *        If the file contains the substring '%chunk', we consider it as a template, the %chunk will be replaced with an integer
 *        See the doc about the constructor, the SetFileName, SetCompression and SetChunkNb for the details.
 */

class CompressedFileName
{
	public:
		CompressedFileName(const string& name="");
		bool operator==( const CompressedFileName& cf) { return GetName()==cf.GetName(); }; // Different compressions, same name => same file !
		const string& GetName() const;
		const string GetCompressionExtension() const { return __compression_extension; };
		CompressionAlgorithm GetCompression()  const { return __compression; };
		void SetCompression ( const string& extension );
		void SetCompression ( CompressionAlgorithm );
		void SetFileName    ( const string& file_name );
		void SetChunkNb     ( size_t chunk );
		operator bool ()  const { return __file_name.size()+__template.size()!=0; };	// if (file_name) { cerr << "file name initialized !" }
	
	private:
		void operator=(const string&) {}; // For security, you MUST do a cast: file_name = (CompressFileName) some_string;
		string __file_name;	              // The name WITHOUT the compression extension, if any
		string __template;                // A template, may be empty (ex. file.%chunk.out)
		string::size_type __chunk_pos;    // The position of %chunk in the template, if npos the template is empty
		string __compression_extension;   // "", "gz", "bz2", etc.
		CompressionAlgorithm __compression;
};

inline ostream & operator<<(ostream& os, const CompressedFileName & cfn) {
	os << cfn.GetName() << '.' << cfn.GetCompressionExtension();
	return os;
}

/* 
 *  All programs share THE SAME Parameters singleton
 *  BUT Each program defines its own object to initialize Parameters
 *  Those ParametersInitializerXXX classes are friends of Parameters, so that they can initialize the private members
 *  See under the ParametersInitializer class for the details
 */

/**
 * \brief this class represents the command line parameters.
 *        It is a singleton, thus it may be used by any object/function through the Instance() static functions
 *        This singleton is initialized by an object of the ParametersInitializer
 *
*/

class Parameters: public Singleton<Parameters> {

private:
	// The constructor is private because of the singleton semantics	
	Parameters();
   
public:
	friend class Singleton<Parameters>;
	friend class ParametersInitializer;
	friend class ParametersInitializerAlign;
	friend class ParametersInitializerSelf;
	friend class ParametersInitializerMap;
	friend class ParametersInitializerMappe;
	friend class ParametersInitializerIndex;
	friend class ParametersInitializerSlice;
	friend class ParametersInitializerStat;
	friend class ParametersInitializerConvert;
	friend class ParametersInitializerFreq;
	friend class ParametersInitializerChain;

	const CompressedFileName& GetOutputFilename()    const { return __results_filename; };
	const CompressedFileName& GetOutputBwdFilename() const { return __results_bwd_filename; };
	const CompressedFileName& GetBankFilename()      const { return __bank_filename; };
	string GetBankName()                             const { return __bank_name;};
	const CompressedFileName& GetQueryFilename()           const { return __query_filename; };
	const CompressedFileName& GetBwdFilename()             const { return __bwd_filename; };
	void SetChunkNb( size_t chunk_nb );
	void SetSilent()  { __silent = true;  __verbose = false; };
	void SetVerbose() { __silent = false; __verbose = true; };

	uint16_t GetIndexNbPass()       const {return __bank_nb_ind_pass;};
    uint16_t GetBlackListMaxCount() const {return __black_list_max_count;};

	bool IsOutputPipe()        const { return !GetOutputFilename(); };
	bool IsSelf()              const { return __self; };
	bool IsToUpper()           const { return __to_upper; };
	bool IsPairEnd()           const { return __pair_end; };
	bool IsePCR()              const { return __epcr; };
	bool IsGenome()            const { return __genome; };
	bool IsSilent()            const { return __silent; };
	bool IsVerbose()           const { return __verbose; };
	bool IsHspOnTheFly()       const { return __is_hsp_on_the_fly; };
	bool IsDynProgramming()    const { return __dyn; };
	bool IsSevereScoringScheme() const { return __strict; };
	bool IsBestScore()         const { return __best_score; };
	bool IsIndexMode()         const { return __index_mode; };
	bool IsSaveIndexMode()     const { return __save_index_mode; };
	bool IsHighThroughputMode() const { return __high_throughput_mode; };
	
	uint32_t GetMaxDistMate()     const { return __maxdistmate; };
	uint8_t GetStep()            const { return __step; };
	uint16_t GetComplexity()      const { return __complexity; };
	uint8_t GetLowerLevel()      const { return __lower_level; };
	uint8_t GetDeepestLevel()    const { return __deepest_level; }; 
	uint64_t GetMaxDistTandem()   const { return __maxdisttandem; };
	uint8_t GetOutputFormat()    const { return __output_format; };
	uint64_t GetOutputChunkSize() const { return __output_chunk_size; };
	uint8_t GetInputFormat()     const { return __input_format;  };
	uint16_t GetCutoffSlide()     const { return __cutoff_slide ; };
	uint16_t GetCutoffDyn()       const { return __cutoff_dyn; };
	uint16_t GetCutoffHsp()       const { return __cutoff_hsp; };
	string GetMskStr()         const { return __msk_string;};
	
	string GetLogLevel()       const { return __log_level; };
	  
	int16_t GetMatchReward()      const { return  __match;       };
	int16_t GetMisMatchPenality() const { return  __mismatch;    };
	int16_t GetNMatchPenality()   const { return  __Nmatch;      }; 
	int16_t GetGapOpenPenality()  const { return  __gapop;       };
	int16_t GetGapExtenPenality() const { return  __gapext;      }; 
	int16_t GetUnGapDropoff()     const { return  __ungapdropoff;};
	int16_t GetGapDropoff()       const { return  __gapdropoff;  };  
	  
	int32_t  GetHspMaxMismatch()  const { return __max_mismatch; };
	uint32_t GetHspMinSize()      const { return __size_min;     };
	uint8_t GetHspMinPIdent()      const { return __ident_min;    };
	int32_t  GetHspMaxGap()       const { return __max_gap;      };
	float GetHspMinRelScore()  const { return __min_ratio_of_max_score; };
	float GetHspMinRelLength() const { return __min_ratio_of_max_length;};
	
	bool IsStatFrequency()  const { return __stat_frequency; };
	bool IsStatDescending() const { return __stat_descending;};
	bool IsStatPositions()  const { return __stat_positions; };
	bool IsStatAll()        const { return __stat_all;       };
	uint8_t GetStatCutoff()   const { return __stat_cutoff;    };
	uint8_t GetStatLevel()    const { return __stat_level;     };
	uint8_t GetNbThreads()    const { return __threads_nb;     };

	// Frequency
	uint8_t 	GetCountCutoff() 		const { return __count_cutoff;};
	size_t 	GetCondemnedMinSize()	const { return __condemned_min_size;};
	double 	GetConseqFreq()			const { return __conseq_freq;};
	double 	GetConseqUnFreq()		const { return __conseq_un_freq;};
	size_t 	GetSuitableMinSize()	const { return __suitable_min_size;};
	size_t 	GetSuitableMaxSize()	const { return __suitable_max_size;};
	size_t 	GetSuitableChunkSize()	const { return __suitable_chunk_size;};
	
	// Chain
	const CompressedFileName& GetGlintOutputFilename()   const { return __glint_output_filename; };
	const CompressedFileName& GetBlastpOutputFilename()  const { return __blastp_output_filename; };
	const CompressedFileName& GetAnchorsOutputFilename() const { return __anchors_output_filename; };

	string GetTargetdir()             const { return __target_dir; };
	string GetChainCutoffs()          const { return __chain_cutoffs; };
	uint32_t  GetMaxOverlap()            const { return __max_overlap; };
	bool   IsRbh()                    const { return __rbh; }

	void Warn(const string& msg="", const string& data="") const;
	
private:

	// Flags   	
    // Parameters modifying the input behaviour 
    bool __pipe;                   // If true the query sequence is provided by stdin
    bool __to_upper;               // If true letters are converted to uppercase
    bool __index_mode;             // If true we'll have to index the bank - If false, we'll have to read the index
    bool __save_index_mode;        // If true we'll save the index (only used if index_mode is true)
     
    // Parameters modifying the output behaviour 
	bool __verbose;				   // If true the program is more verbose
	bool __silent;				   // If true the program is silent
	bool __is_hsp_on_the_fly;        // If true the hsp are displayed once they are computed
		
	//Comparison mode
	bool __genome;                 // If true the query genome is seen as a Bank (should have been previously indexed by glintindex)
	bool __self;                   // If true the genome is compared against itself
	bool __dyn;                    // If false do not perform a dynamic programming (much faster but less sensitive)
	
	// high-throughput mode
	bool __high_throughput_mode;   // high throughput mode
	uint8_t __threads_nb;            // Number of threads (default = 1)
	bool __pair_end;               // If true we are mapping pairends against a genome
	uint16_t __maxdistmate;           // Maximum distance between mate-pair sequences
	bool __epcr;                   // If true we are mapping primers against a genome	

	// Controling the aligment process
	bool __strict;                 // If true a severe scoring sheme is used for the alignment (dyn prog) see scoring.cpp
	
	//For pair-end mapping		    	
	uint8_t __step;                  // Controlling the number of nucleotides to be shifted for each window move -default => 1)
	uint16_t __complexity;            // The complexity algorithm used for indexation or search (0/1/2)	
	uint8_t __lower_level;
	uint8_t __deepest_level; 
	uint16_t __cutoff_slide;          // minimum score to perform an ungapped alignment
	uint16_t __cutoff_dyn;			   // minimum score to perform an gapped alignment
	uint16_t __cutoff_hsp;			   // minimum score to report an hsp
				
	uint64_t __maxdisttandem;           // Maximum distance between seeds for self comparison dedicated to tandem repeat detection
		
	// Controlling the scoring scheme	
	int16_t __match;                    // Reward for a match 
    int16_t __mismatch;                 // Reward (penality) for a mismatch
    int16_t __Nmatch;                   // Reward (penality) for a matching with an N
    int16_t __gapop;                    // Gap opening penality
    int16_t __gapext;                   // Gap extension penality
    int16_t __ungapdropoff;             // Dropoff value for ungapped alignment
    int16_t __gapdropoff;               // Dropoff value for gapped alignment
	
	// Controlling the hsp to be returned by the program
	bool __best_score;	             // Only best hsps are returned   
    int32_t __max_mismatch;             // Maximum allowed mismatch to record an hsp (default = -1 ==> no check)
    int32_t __max_gap;                  // Maximum allowed gap to record an hsp (default = -1 ==> no check)
    uint32_t __size_min;                // Minimum allowed size of an hsp to record it (default = 0)
    uint8_t  __ident_min;             // Minimum allowed identity of an hsp to record it (default = 0)
    float __min_ratio_of_max_score;  // Minimum score, expressed in percentage of max possible score for the query sequence, to record an hsp (default = -1 ==> no check)
    float __min_ratio_of_max_length; // Minimum length, expressed in percentage of query sequence length, to record an hsp (default = -1 ==> no check)

    uint8_t __output_format;
    uint64_t __output_chunk_size;     // The number of sequences written in the output file (only used by glintconvert for now)
    uint8_t __input_format;
    
    mutable CompressedFileName __results_filename;
	mutable CompressedFileName __results_bwd_filename;
	
    CompressedFileName __query_filename;
    CompressedFileName __bank_filename;
    CompressedFileName __bwd_filename;
    string __log_level;
    
	// guessing compression from a file name, and changing the file name
////	CompressionAlgorithm __GuessCompression( string&);

	// Patch to remove compression extension to the __bwd_filename in fast(a|q) pair-end mode
////	void __PATCH_BwdCompressionExtension( CompressionAlgorithm& , string& );
	
	// glintindex	
	string __bank_name;
    int    __bank_nb_ind_pass;     // number of passes in the indexation process
    uint16_t  __black_list_max_count; // Maximum allowed occurence of a word, otherwise => black list
    string __msk_string;           // The mask, as a string

	// glintfreq
	uint8_t 	__count_cutoff;
	size_t 	__condemned_min_size;
	double 	__conseq_freq;	
	double	__conseq_un_freq;
	size_t 	__suitable_min_size;
	size_t 	__suitable_max_size;
	size_t 	__suitable_chunk_size;
	
	// Used by the Warn function
	mutable map<string,int>             __warnings;
	mutable map<string,vector<string> > __warnings_data;
	
	// glintstat
	bool  __stat_frequency;
	bool  __stat_descending;
	bool  __stat_positions;
	bool  __stat_all;
	uint8_t __stat_cutoff;          // Do not display frequency of words if nb of pos < stat_cutoff
	uint8_t __stat_level;

	// glintchain
	CompressedFileName __glint_output_filename;
	CompressedFileName __blastp_output_filename;
	CompressedFileName __anchors_output_filename;
	
	string __target_dir;
	string __chain_cutoffs;
	uint32_t __max_overlap;
	bool __rbh;
	
	void __InitBankName();

	bool __FindIndex();			   // Return true if the index is found
	void __LogParameters();		   // For debugging purpose only
	
	void __ParseFileName(string&,string&);
	int  __argc;
	void (*__Help)();
};

/**
 * \brief This functor initializes the Parameters singleton from command line
 *        this is a hierarchy of classes, as each program of the suite has a different initialization algorithm
 *        Init calls the virtual functions _Init and _Validation, defined in the subclasses
 * \code
 * \pre The Parameters and CompressionInfo singletons must be already initialized
 * Parameters& parameters = Parameters::Instance();
 * ParametersInitializeSubClass prm_init(argc,argv,Usage);
 * \endcode
*/
class Command;
class ParametersInitializer {

public:
	ParametersInitializer() : p(Parameters::Instance()),compression_info(CompressionInfo::Instance()){};
	void Init(int argc,char* argv[],const Command& cmd); 

protected:
    // _Init starts initialization and returns a CSimpleOpt
    // _IniT2 terminats initialization, starting from the returned CSimpleOpt
	virtual CSimpleOpt _Init (int, char* argv[], const Command&) = 0;
	virtual void _Init2(const CSimpleOpt&, const Command&) = 0;

	bool _CommonValidation();
	virtual bool _Validation() = 0;
	Parameters& p;
	const CompressionInfo& compression_info;
};

/*
 * This enumeration is used by ALL ParametersInitializer classes 
 */
enum {
	OPT_ANCHORSOUT,
	OPT_BANK_NAME,
	OPT_BLASTPOUT,
	OPT_COMPLEXITY,
	OPT_LC_FILTERING,
	OPT_CONDEMNED_MIN_SIZE,
	OPT_CONSEQFREQ,
	OPT_CONSEQUNFREQ,
	OPT_COUNT_CUTOFF,
	OPT_CUTOFF_DYN,
	OPT_CUTOFF_HSP,
	OPT_CUTOFFS,
	OPT_CUTOFF_SLIDE,
	OPT_DEEPEST_LEVEL,
	OPT_EPCR,
	OPT_FILL,
	OPT_GAP_OPENING,
	OPT_GAP_EXTENSION,	
	OPT_GENOME,
	OPT_GLINTOUT,
	OPT_HELP,
	OPT_HIGHTHROUGHPUT,
	OPT_IDENT_MIN,
	OPT_INPUTFORMAT,
	OPT_LENGTH_RATIO_MIN,
	OPT_LOG_LEVEL,
	OPT_LOWER_LEVEL,
	OPT_MATCH,	
	OPT_MASK,
	OPT_MAX_COUNT,
	OPT_MAX_DISTMATE,
	OPT_MAX_GAP,
	OPT_MAX_MIS,
	OPT_MAXOVERLAP,
	OPT_MISMATCH,
	OPT_NB_PASS,
	OPT_NOGAP,
	OPT_NORBH,
	OPT_OUTPUTFORMAT,
	OPT_OUTPUTSTREAM,
	OPT_PAIREND,
	OPT_PIPE,
	OPT_RESULTSFILENAME,
	OPT_SCORE_RATIO_MIN,
	OPT_SELF,
	OPT_TOUPPER,
	OPT_SILENT,
	OPT_SIZE_MIN,
	OPT_BEST,
	OPT_SLICESIZE,
	OPT_STEP,
	OPT_STRICT,
	OPT_SUITABLE_CHUNK_SIZE,
	OPT_SUITABLE_MAX_SIZE,
	OPT_SUITABLE_MIN_SIZE,
	OPT_TANDEM,
	OPT_TARGETDIR,
    OPT_VERBOSE,
    OPT_STAT_FREQUENCY,
    OPT_STAT_DESCENDING,
    OPT_STAT_POSITIONS,
    OPT_STAT_ALL,
    OPT_STAT_CUTOFF,
    OPT_STAT_LEVEL,
    OPT_COMPRESS,
    OPT_SAVE_INDEX,
    OPT_A,
    OPT_CHUNKSIZE
};

#endif

