/*
   $Id: ComparisonFrequency.h 648 2011-10-31 12:59:22Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __SUITABLEREGION_H__
#define __SUITABLEREGION_H__

#include <memory>
using namespace std;
#include <stack>

enum FREQ
	{
	LOW = 0,
	HIGH = 1,
	CONDEMNED = 2
	};

typedef unsigned char freqCode;

#include "Parameters.h"
#include "QueryBuffer.h"
#include "Display.h"

/**
 * \brief The SuitableRegion class is dedicated to the identification of suitable regions for probe
 *        deesign. 
 *       - A suitable region is a region where the proposition of repeats is low :
 *           - first all regions with a high proportion of repeats are masked ( __SentencingRegions )
 *           - second, large regions with an excess of unfrequent word are selected ( __SelectingSuitableRegions )
 *       - The object essentially keeps a pointer on the query sequence and a FrequentBuffer (keeping track of the repeats, 
 *         lower case letters in the sequence)
*/
class SuitableRegion
{
	public :
		SuitableRegion();
		void Run();
				
	protected:
		Parameters&  _parameters;
		QueryBuffer* _query_ptr;
		DisplayFreq _display;
		
		QueryBuffer* _InitQueryBuffer() { return new QueryBufferFasta( _parameters.GetQueryFilename().GetName(), _parameters.GetQueryFilename().GetCompression());}; 
	
		const vector<regionFreq>&  _GetSuitableRegions()  {return __suitableRegions;};			
		QueryBuffer* _GetQueryPtr() const { return _query_ptr;};
			
		//The methods for the __frequentCodeBuffer
		const Buffer&  _GetFrequentBuffer()           { return __frequentCodeBuffer; };	
		bool           _FrequentBufferPut(freqCode f) { return __frequentCodeBuffer.Put((char) f); };
		size_t         _GetBufferSize()               { return __frequentCodeBuffer.size(); };
		freqCode*      _GetBufferBase()               { return (freqCode*) __frequentCodeBuffer.GetBase(); };
		void           _ClearFrequentBuffer()         { __frequentCodeBuffer.Clear(); };
					
	private:	
				
		void __MaskRepeats();
		void __DetectSuitableRegions();	
		void __SentencingRegions();
		void __SelectingSuitableRegions();
		void __MaskSentencedRegions();
		 
		void __AddSuitableRegion( uint32_t start, uint32_t end ); 
		 	
		double __SentenceScore[3];
		double __SuitableScore[3];

		Buffer  __frequentCodeBuffer;     // Buffer containing the recoded sequence : LOW, HIGH or CONDEMNED (see enum FREQ)

		vector<regionFreq>  __sentencedRegions;
		vector<regionFreq>  __suitableRegions;	
	
	//For debugging only	
	protected :			
		void _DisplayFrequentBuffer();	
};

inline uint32_t __max0( int v )
{
	return ( v > 0 ) ? v : 0;
}

//Test if a char (nucleotide) is in lower case
inline bool IsLower(const unsigned char c) { return ( (c&0x20) != 0) ? true : false; }

inline unsigned char Decode(freqCode f)
{
	unsigned char c;
	
	switch ( f ) 
	{
		case HIGH      : c = 'H'; break;
		case LOW       : c = 'L'; break;
		case CONDEMNED : c = 'C'; break;
		default        : c = '?'; break;
	}
	
	return c;
}


#endif
