/*
   $Id: alignment.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __ALIGN_H__
#define __ALIGN_H__

#include <stdio.h>
#include <stdlib.h>
#include "ptypes.h"
#include "scoring.h"
#include <algorithm>
#include <list>

#include "hsp.h"

enum {NONE, DIAG, LEFT, ABOVE, OPEN_Q, OPEN_S, CONT_GQ, CONT_GS, DONE};

enum Direction { UPSTREAM = -1, DOWNSTREAM = 1 };

/*
    The objects related to the alignment
        Associations:
            AlignDirection     <---> AlignEngine
            AlignEngine ---> QueryBuffer
            AlignEngine ---> Alignconstants

        Dependencies:
             AlignDirection     <---> AlignEngine

*/

struct Interval {
	uint16_t min;
	uint16_t max;	
};

class match;
class ComparisonBase;
class QueryBuffer;
class CompactTree;

/**
 * \brief The alignment algorithms (including gapped alignment using dynamic programming) are implemented here
 *        The main entry point is the Align function, who performs an ungapped, then gapped align
 */

class AlignEngine {
public:
    AlignEngine(int st); 
	bool Align(match*, Hsp&, DiagInt&, const QueryBuffer&);

    uint16_t GetScore()        const { return _a_score;};
	uint32_t GetQueryStart()   const { return _q_start;};
    uint32_t GetQueryEnd()     const { return _q_end;};
	uint32_t GetSubjectStart() const { return _s_start;};
    uint32_t GetSubjectEnd()   const { return _s_end;};

private:

    uint32_t __LocalAlign (Hsp&, const QueryBuffer&);
	void  __InitDPparameters(const QueryBuffer&);
	void __UpdateAlignementStartEnd();
    
    uint16_t __RecordHspScript(Hsp&, const QueryBuffer&) const;
    int   __RecordAlignment(Hsp&);
    bool    __ValidHsp (Hsp&, const QueryBuffer& );
	DiagInt __ToHsp    (Hsp&, const QueryBuffer&); 
   
	virtual void __GetSeqInfo(Hsp&, const QueryBuffer&);
    virtual int  __LookRight(uint32_t, uint32_t, const QueryBuffer&);
    virtual int  __LookLeft(uint32_t, uint32_t, const QueryBuffer&);

    void show_score_matrix(int , int , int , int , FILE *, const QueryBuffer&);
    
#ifdef DEBUGLOG
	void LogScore(const char* log_type);
	void LogAlign(const char* log_type, const char* header);
	void show_score_matrix(int l_inf, int l_max, int c_inf, int c_max,const char* log_type,const QueryBuffer&);
#endif	
	
	const int8_t __strand;
	const uint8_t __strand01;
	
protected:
    
    const Parameters& __parameters;
    const Bank& __bank;
    const Mask& __mask; 
    
    
    void __SetDirection(enum Direction dir) { direction=dir; };
	int __GetDPMatrixQueryMove() const { return ( direction > 0 ) ? 1 : -1; };
	int __GetDPMatrixSubjectMove() const { return ( direction == __strand ) ? 1 : -1; };
    uint32_t __GetDPMatrixQueryStart() const {return ( direction > 0 ) ? _q_end + 1 : _q_start - 1;}; 
    uint32_t __GetDPMatrixSubjectStart() const { return ( direction == __strand ) ? _s_end + 1 : _s_start -1;};
    int8_t _GetStrand() const { return __strand; }; 
	uint8_t _GetStrand01() const { return __strand01; }; 

    int _ComputeDPMatrixLine();
    virtual bool _ValidDPMatrixline( );
         
    // For dynamic programming
    // The two DP matrix indices
    uint16_t I;
	uint16_t J;
	
    // The 3 scoring matrices
	int16_t S[DYNPROG_MAT_SIZE+1][DYNPROG_MAT_SIZE+1];
	int16_t F[DYNPROG_MAT_SIZE+1];
	int E;
		
	// The 3 trace matrices for backtracking
	uint8_t Trace   [DYNPROG_MAT_SIZE+1][DYNPROG_MAT_SIZE+1];
	uint8_t Trace_E [DYNPROG_MAT_SIZE+1][DYNPROG_MAT_SIZE+1];
	uint8_t Trace_F [DYNPROG_MAT_SIZE+1][DYNPROG_MAT_SIZE+1];
	
	
    void __InitScoreMatrix( );
    void __InitTraceMatrix( );
    
	int16_t *sdiag, *scurrent, *sleft, *sabove, *fabove;
	const unsigned char *nQuery, *nSubject;
	
	// Storing best scores and best cells	
	int32_t _bestscore;
	int dropoff;	
	
	enum Direction direction;		
	uint16_t bestSi;
	uint16_t bestSj;
	uint32_t bestSi_left;
	uint32_t bestSi_right;
	uint32_t bestSj_left;
	uint32_t bestSj_right;
	
	uint32_t _q_start;
	uint32_t _q_end;
	uint32_t _s_start;
	uint32_t _s_end;
		
	uint32_t _a_score;
	uint32_t _a_size;
	
	uint64_t max_subject_size;
	uint64_t max_query_size;
	
	int max3(int a, int b, int c);
	int max2_E(int a, int b);
	int max2_F(int a, int b);
	
	Interval Jrange;

    const AlignConstants& scoring;
    //AlignDirection align_direction;

    const CompactTree& ctree;

    // Used by _RecordHspScript (stores the Watson-Crick relation A <->T etc.)
    int Rev[256];
    
};

/*
      Some utility functions:

		Functions defined for the purpose of dynamic programming and more specifically for the traceback (which cell realized the maximum)

*/

inline int AlignEngine::max3(int a, int b, int c)
{
	int max;

	if ( a > b ) {
		max = a;
		Trace[I][J]=DIAG;
	} else {
		max=b;
		Trace[I][J]=LEFT;
	}
	if ( max < c ) {
		max=c;
		Trace[I][J]=ABOVE;
	}

	return max;
}

inline int AlignEngine::max2_E(int a, int b)
{
	int max;

	if ( a > b ) {
		max = a;
		Trace_E[I][J]=OPEN_Q;
	} else {
		max=b;
		Trace_E[I][J]=CONT_GQ;
	}

	return max;
}

inline int AlignEngine::max2_F(int a, int b)
{
	int max;

	if ( a > b ) {
		max = a;
		Trace_F[I][J]=OPEN_S;
	} else {
		max=b;
		Trace_F[I][J]=CONT_GS;
	}

	return max;
}


class AlignEngineSelf: public AlignEngine {
public:	
	AlignEngineSelf(int st): AlignEngine(st) {};
private:
    virtual int  __LookRight(uint32_t, uint32_t, const QueryBuffer&);
    virtual int  __LookLeft (uint32_t, uint32_t, const QueryBuffer&);
    virtual void __GetSeqInfo(Hsp&, const QueryBuffer& );
    virtual bool _ValidDPMatrixline( );
    
};

#endif
