/*
   $Id: buffer.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef __BUFFER_H__
#define __BUFFER_H__

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <memory>
using namespace std;

#include "ptypes.h"

/**
 * \brief Buffer is a low-level buffer, used for storing sequences, binary indexes, etc.
 *        - You may specify a minimum size (default 500000), a maximal size (default 0, ie no maximal size)
 *        - The functions Putc adjust the buffer size when necessary (and when possible)
 *        - IsFull returns true only if the buffer cannot be adjusted any more (the max size has been reached)
 *        - The memory may be managed by Buffer or not, using malloc/free. See the constructors for the details
 */

class Buffer
{
    public:
	    Buffer(size_t min=500000,size_t max=0);
	    Buffer(void* begin, size_t size);
	    Buffer(const Buffer&);
	    ~Buffer() {if (begin!=NULL && owner) free(begin);};
	    bool Put(char);
	    bool Put(uint16_t);
	    bool Put(uint32_t);
		bool Put(uint64_t);
		bool Put(void*, size_t);
		// Warning ! If  not owner, Clear and Reverse do nothing but say nothing
	    void Clear()   { if(owner) end = begin; };
	    void Reverse();
	    void* GetBase()  const { return begin;};
		void* GetBegin() const { return begin;};
		void* GetEnd()   const { return end;}
	    size_t capacity() const {return cur_capacity;};
	    size_t size()  const {return (char*)end - (char*)begin;};
	    bool IsFull()  const {return size()==capacity() && !__Realloc();};   // The array is full if we cannot reallocate memory
	    bool IsEmpty() const {return begin==end;};
	
		void SetBegin(void *nbegin) {begin=nbegin;};
		void SetEnd(void *nend) {end=nend;};
		void ToUpper();    // Convert case of characters in the buffer
		void ToLower();    // Convert case of characters in the buffer

    private:
	    void operator=(const Buffer&){};	// Removing the operator= !
	    size_t min_capacity;
	    size_t max_capacity;
	    mutable size_t cur_capacity;
	
	    mutable void* begin;    // points to the first character of the array
	    mutable void* end;      // points AFTER the LAST character of the array
	
	    bool owner;
	    bool __Realloc() const throw(bad_alloc);
};

/*
   A Buffer_ptr is a smart pointer, it will be useful to transfer the responsability of fa_bfr and bi_bfr between
   BankBuffer and Bank
*/

typedef auto_ptr<Buffer> Buffer_ptr;

/* The inline functions of the class Buffer
*/
/**
 * \brief The Put functions (several versions)
 *        Performance trick: 
 *          - inline functions
 *          - write the most often used instruction just after the if
 * \param [in] c The data (may be char, uint16_t, Unit4, uint64_t) to put inside the buffer
 * \return true if OK, false if there is a capacity issue
*/

inline bool Buffer::Put(char c)
{
    if (size()+sizeof(char) <= cur_capacity)
    {
        char* tmp = (char*)end;
        *(tmp++)=c;
        end = (void*)tmp;
        return true;
    }
    else
    {
        if (__Realloc())
        {
            char* tmp = (char*)end;
            *(tmp++)=c;
            end = (void*)tmp;
            return true;
        }
        else
            return false;
    }
}

inline bool Buffer::Put(uint16_t i)
{
    if (size()+sizeof(uint16_t) <= cur_capacity)
    {
        uint16_t* tmp = (uint16_t*)end;
        *(tmp++)=i;
        end = (void*)tmp;
        return true;
    }
    else
    {
        if (__Realloc())
        {
            uint16_t* tmp = (uint16_t*)end;
            *(tmp++)=i;
            end = (void*)tmp;
            return true;
        }
        else
            return false;
    }
}

inline bool Buffer::Put(uint32_t i)
{
    if (size()+sizeof(uint32_t) <= cur_capacity)
    {
        uint32_t* tmp = (uint32_t*)end;
        *(tmp++)=i;
        end = (void*)tmp;
        return true;
    }
    else
    {
        if (__Realloc())
        {
            uint32_t* tmp = (uint32_t*)end;
            *(tmp++)=i;
            end = (void*)tmp;
            return true;
        }
        else
            return false;
    }
}

inline bool Buffer::Put(uint64_t i)
{
    if (size()+sizeof(uint64_t) <= cur_capacity)
    {
        uint64_t* tmp = (uint64_t*)end;
        *(tmp++)=i;
        end = (void*)tmp;
        return true;
    }
    else
    {
        if (__Realloc())
        {
            uint64_t* tmp = (uint64_t*)end;
            *(tmp++)=i;
            end = (void*)tmp;
            return true;
        }
        else
            return false;
    }
}

/**
 * \brief The Put function (memory area version)
 *        Resize the buffer if necessary, then call memcpy
 * \param [in] src The data base address (void*)
 * \param [in] l   The length
 * \return true if OK, false if there is a capacity issue
*/
inline bool Buffer::Put(void* src, size_t l)
{
	bool rvl = true;
	if (size()+l<=cur_capacity || __Realloc())
	{
		memcpy(end,src,l); 
		end = (void*) ((char*) end + l);
	}
	else
	{
		rvl = false;
	}
	return rvl;
}
#endif
